﻿// Unicessing Buildings4 サンプル
// Written by NISHIDA Ryota, http://dev.eyln.com

using UnityEngine;
using UnityEngine.VR;
using Unicessing;

// Oculus Touchを上下に動かすとビルを伸ばしたり縮めたりできます。
// Aボタン(Fire1)を押しながら右手を伸ばした方向に太陽を置けます。
// Bボタン(Fire2)を押しながら両手を近づけたり遠ざけるとビルの間隔を変えられます。
public class UnicessingBuildings4 : UGraphics
{
    public Transform directionaLight;  // 太陽
    public float buildingStep = 15.0f; // ビルの配置間隔
    public Vector3 buildingSize = new Vector3(5, 100, 5);  // ビルの大きさ
    public float floorHeight = -2.0f;

    Vector3 leftPos, rightPos;         // 左手、右手コントローラーの位置
    Quaternion leftRot, rightRot;      // 左手、右手コントローラーの姿勢

    // 初期化
    protected override void Setup()
    {
        rotateDegrees();     // 回転の単位をラジアンではなく「度」にする
        stroke(0, 255, 100); // 緑色のワイヤーフレームにする
    }

    // 毎フレームの更新と描画
    protected override void Draw()
    {
        UpdateHands();    // 手の位置を更新
        DrawHands();      // 手の位置に箱を描画
        DrawBuildings();  // ビルを描画
    }

    // 手の位置を更新
    void UpdateHands()
    {
        leftPos = InputTracking.GetLocalPosition(VRNode.LeftHand);   // 左手位置
        rightPos = InputTracking.GetLocalPosition(VRNode.RightHand); // 右手位置
        leftRot = InputTracking.GetLocalRotation(VRNode.LeftHand);   // 左手姿勢
        rightRot = InputTracking.GetLocalRotation(VRNode.RightHand); // 右手姿勢

        if (isButton("Fire1")) // Oculus TouchのAボタンを押している間
        {
            // 右手を向ける位置にあわせて太陽の位置を変更
            Vector3 headPos = InputTracking.GetLocalPosition(VRNode.CenterEye); // 頭の位置
            Vector3 bodyPos = headPos + new Vector3(0, -0.2f, 0); // 頭の位置より少し下を体とみなす
            directionaLight.position = Vector3.zero;
            directionaLight.LookAt(bodyPos - rightPos);           // 右手がある方向の逆を向く
        }

        if (isButton("Fire2")) // Oculus TouchのBボタンを押している間
        {
            // 手と手の近さにあわせてビルの間隔を変更
            buildingStep = buildingSize.x + (leftPos - rightPos).magnitude * 10.0f * buildingSize.x;
        }
    }

    // 手の位置に箱を描画
    void DrawHands()
    {
        float handSize = 0.15f;
        DrawHandBox(leftPos, leftRot, handSize);
        DrawHandBox(rightPos, rightRot, handSize);
    }

    // 手の箱を描画
    void DrawHandBox(Vector3 pos, Quaternion quat, float size)
    {
        pushMatrix();   // 現在位置、姿勢などを保存
        translate(pos); // 移動する
        rotate(quat.eulerAngles); // 回転する
        //uniTransfrom.rotation = quat; // rotate(quat.eulerAngles)のかわりにこうしてもよい
        noFill();       // 塗りつぶさない
        box(size);      // 箱を描く
        popMatrix();    // 現在位置、姿勢などを戻す
    }

    // ビルを描画
    void DrawBuildings()
    {
        // 太陽の位置が上のときは塗りつぶし、下のときはワイヤーフレームのみにする
        if (directionaLight.forward.y < 0.2f) fill(20, 100, 20);
        else noFill();

        randomSeed(0); // 乱数系列を一定にして各ビルが毎回同じ高さになるようにする

        // ビルを描画
        pushMatrix(); // 現在位置、姿勢などを保存
        translate(buildingStep * 0.5f, floorHeight, buildingStep * 0.5f); // ビル全体の位置を調節
        for (int z = -5; z < 10; z++)
        {
            for (int x = -5; x < 5; x++)
            {
                pushMatrix();
                float handY = ((x + z) % 2 == 0) ? leftPos.y : rightPos.y; // 左手右手の振り分け
                float h = buildingSize.y * random(0.1f, handY + 1);        // ビルの高さ
                translate(x * buildingStep, h * 0.5f, z * buildingStep);
                box(buildingSize.x, h, buildingSize.z); // ビルを箱として描画
                popMatrix();
            }
        }
        popMatrix();  // 現在位置、姿勢などを戻す
    }
}
